'use strict';

/* jshint node: true, esnext: true, loopfunc: true, undef: true, unused: true */

/*
 * @author: 绝云（wensen.lws）
 * @description: cron expression parser/stringifier
 * @note: accepted format
 *   (*) * * * * *
 *    |  | | | | |
 *    |  | | | | --- day of week       (1 - 7 )
 *    |  | | | ----- month             (1 - 12)
 *    |  | | ------- day of month      (1 - 31)
 *    |  | --------- hour              (0 - 23)
 *    |  ----------- minute            (0 - 59)
 *    -------------- (optional) second (0 - 59)
 * @note: current supported expression:
 *   - *       (every)
 *   - 0/15    (interval)
 *   - 1-5     (range)
 *   - 1,2,4,5 (select)
 *   - L       (last day of month)
 */

var lang = require('zero-lang');
var sprintf = require('zero-fmt/sprintf');

var NAMES = ['second', 'minute', 'hour', 'dayOfMonth', 'month', 'dayOfWeek'];
var RANGE = {
    second: [0, 59],
    minute: [0, 59],
    hour: [0, 23],
    dayOfMonth: [1, 31],
    month: [1, 12],
    dayOfWeek: [1, 7]
};
var ALIAS = {
    dayOfWeek: ['SUN', 'MON', 'TUE', 'WED', 'THU', 'FRI', 'SAT', 'SUN'],
    month: ['DEC', 'JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC']
};

function alias2number(alias, name) {
    var errorStr = sprintf('Invalid str: %s', alias);
    alias = lang.uc(alias);
    if (lang.isNaN(alias) && name !== 'month' && name !== 'dayOfWeek') {
        throw errorStr;
    }
    if (lang.isNaN(alias) && lang.indexOf(ALIAS[name], alias) < 0) {
        throw errorStr;
    }
    return lang.isNaN(alias) ? lang.indexOf(ALIAS[name], alias) : lang.toInteger(alias);
}
function number2alias(number, name) {
    var aliasArr = ALIAS[name];
    if (!aliasArr) {
        return number;
    }
    var alias = aliasArr[number];
    return alias ? alias : number;
}

module.exports = {
    alias: ALIAS,
    names: NAMES,
    range: RANGE,
    alias2number: alias2number,
    number2alias: number2alias,
    parse: function parse(expression) {
        var parts = expression.split(' ');
        var length = parts.length;
        var partNames = lang.clone(NAMES);
        var cronObj = {};

        // check and pre-process
        switch (length) {
            case 5:
                lang.remove(partNames, 0);
                break;
            case 6:
                break;
            default:
                throw sprintf('Invalid cron expression: %s', expression);
        }

        // parse
        function parseRange(atom, name) {
            var fromTo = atom.split('-');
            var from = alias2number(fromTo[0], name);
            var to = alias2number(fromTo[1], name);
            var minMax = RANGE[name];
            var min = minMax[0];
            var max = minMax[1];

            if (!lang.isInRange(from, min, max) || !lang.isInRange(to, from, max)) {
                throw sprintf('%s: %s should be in range %s-%s', name, atom, min, max);
            }
            cronObj[name].range = {
                from: from,
                to: to
            };
            for (var i = from; i <= to; i++) {
                cronObj[name].all.push(i);
            }
            cronObj[name].all = lang.uniq(cronObj[name].all);
        }
        function parseInterval(atom, name) {
            var fromStep = atom.split('/');
            var from = lang.isNumber(fromStep[0]) ? lang.toInteger([0]) : 0;
            var step = lang.toInteger(fromStep[1]);
            cronObj[name].interval = {
                from: from,
                step: step
            };
            for (var i = from; i < RANGE[name][1]; i++) {
                if (i % step === 0 && i >= RANGE[name][0]) {
                    cronObj[name].all.push(i);
                }
            }
            cronObj[name].all = lang.uniq(cronObj[name].all);
        }
        function parseSingle(atom, name) {
            atom = alias2number(atom, name);
            var minMax = RANGE[name],
                min = minMax[0],
                max = minMax[1];
            if (!lang.isInRange(atom, min, max)) {
                throw sprintf('%s: %s should be in range %s-%s', name, atom, min, max);
            }
            cronObj[name].select.push(atom);
            cronObj[name].all.push(atom);
            cronObj[name].select = lang.uniq(cronObj[name].select);
            cronObj[name].all = lang.uniq(cronObj[name].all);
        }
        function parseLastDayOfMonth() {
            cronObj.dayOfMonth.select.push('L');
            cronObj.dayOfMonth.all.push('L');
            cronObj.dayOfMonth.select = lang.uniq(cronObj.dayOfMonth.select);
            cronObj.dayOfMonth.all = lang.uniq(cronObj.dayOfMonth.all);
        }
        function parse(part, name) {
            if (part === '*') {
                return lang.extend(cronObj[name], {
                    every: true,
                    range: {},
                    interval: {},
                    select: [],
                    all: ['*']
                });
            }
            lang.each(part.split(','), function (atom) {
                if (atom.indexOf('-') >= 0) {
                    parseRange(atom, name);
                } else if (atom.indexOf('/') >= 0) {
                    parseInterval(atom, name);
                } else if (atom === 'l' || atom === 'L' && name === 'dayOfMonth') {
                    parseLastDayOfMonth(atom);
                } else {
                    parseSingle(atom, name);
                }
            });
        }

        // process
        lang.each(partNames, function (name, index) {
            cronObj[name] = {
                every: false,
                range: {},
                interval: {},
                select: [],
                all: []
            };
            parse(parts[index], name);
        });
        return cronObj;
    },
    stringify: function stringify(cronObj) {
        var resultStr = '';
        lang.each(NAMES, function (name) {
            var part = cronObj[name];
            var partStr = '';
            var parts = [];
            if (!part) {
                return;
            }
            if (part.every) {
                resultStr += '*';
            } else {
                if (lang.isNumber(part.range.from) && lang.isNumber(part.range.to) && !lang.isNaN(part.range.from) && !lang.isNaN(part.range.to)) {
                    parts.push(sprintf('%d-%d', part.range.from, part.range.to));
                }
                if (lang.isNumber(part.interval.from) && lang.isNumber(part.interval.step) && !lang.isNaN(part.interval.from) && !lang.isNaN(part.interval.step)) {
                    parts.push(sprintf('%d/%d', part.interval.from, part.interval.step));
                }
                if (part.select.length) {
                    parts.push(part.select.join(','));
                }
                partStr = parts.join(',');
                if (lang.trim(partStr) === '') {
                    partStr = '*';
                }
                resultStr += partStr;
            }
            resultStr += ' ';
        });
        return lang.trim(resultStr);
    }
};